# -*- coding: utf-8 -*-
"""
@Author : zhang.yonggang
@File   : dlt_filter.py
@Project: pydlt
@Time   : 2023-10-25 15:28:03
@Desc   : The file is ...
@Version: v1.0
"""
import os.path

import json

from pydlt_plus import DltFileReader


def deal_output(msg, output_content, type_code, f_word=None):
    msg = msg.strip()
    if type_code == 0:
        output_content.append(msg)
    elif type_code == 1 and f_word:
        if f_word in output_content:
            output_content[f_word].append(msg)
        else:
            output_content[f_word] = [msg]
    elif type_code == 2:
        try:
            with open(output_content, 'a', encoding='utf-8') as file:
                file.write(f"{msg} {os.linesep}")
        except IOError:
            print('cannot open file {}'.format(output_content))
    return output_content


class DltFileFilter(DltFileReader):
    """
    A class to read and filter DLT message from DLT file BASE ON DltFileReader.
    Examples::
        # create reader as context manager
        with DltFileFilter("filepath") as reader:
            # read 1 message from file
            message = reader.read_message()

            # read all messages from file
            messages = reader.read_messages()

        # create reader as iterator
        for message in DltFileFilter("filepath"):  # read all messages
            # handle each message
        # filter info
        dff = DltFileFilter("log_2020-01-01_00.00.10.4.dlt", "utf8")
        dff.read_file(["abc", "神奈川県"], output_type="aaaa.txt")
        # translate dlt to other types
        dff.read_file(output_type="aaaa.txt")
    """
    def __init__(self, dlt_file_path, encoding="utf8"):
        if not os.path.exists(dlt_file_path):
            raise FileExistsError
        super(DltFileFilter, self).__init__(dlt_file_path, encoding)

    def read_file(self, filter_info=None, output_type="list"):
        """
        read and filter DLT message
        :param filter_info: filter_info could be a list or string or None(Not input).
        :param output_type: output_type could be a list or text(.txt or .log) or dict.
        :return:
        """
        if isinstance(filter_info, str):
            filter_info = [filter_info]
        # type_code: 0 is list, 1 is json file, 2 is text
        if output_type.lower() == "list":
            output_content = []
            type_code = 0
        elif output_type.lower().endswith(".json"):
            output_content = {}
            type_code = 1
        elif output_type.lower().endswith(".txt") or output_type.lower().endswith(".log"):
            type_code = 2
            output_content = output_type
        for msg in self.__iter__():
            msg = str(msg)
            if filter_info:
                for f_word in filter_info:
                    if f_word in msg:
                        output_content = deal_output(msg, output_content, type_code, f_word)
            else:
                output_content = deal_output(msg, output_content, type_code)
        if type_code == 1:
            with open(output_type, 'w', encoding='utf-8') as json_file:
                json.dump(output_content, json_file, ensure_ascii=False, indent=4)
        return output_content



# Import all classes in the sub modules of pydlt
# F401 is ignored because they will be used from not here but a user of the library
from pydlt_plus.file import DltFileReader, DltFileWriter  # noqa: F401
from pydlt_plus.dlt_filter import DltFileFilter
from pydlt_plus.header import (  # noqa: F401
    ExtendedHeader,
    MessageBusInfo,
    MessageControlInfo,
    MessageLogInfo,
    MessageTraceInfo,
    MessageType,
    MessageTypeInfo,
    StandardHeader,
    StorageHeader,
)
from pydlt_plus.message import DltMessage  # noqa: F401
from pydlt_plus.payload import (  # noqa: F401
    Argument,
    ArgumentBool,
    ArgumentFloat32,
    ArgumentFloat64,
    ArgumentRaw,
    ArgumentSInt8,
    ArgumentSInt16,
    ArgumentSInt32,
    ArgumentSInt64,
    ArgumentString,
    ArgumentUInt8,
    ArgumentUInt16,
    ArgumentUInt32,
    ArgumentUInt64,
    NonVerbosePayload,
    Payload,
    VerbosePayload,
)



"""Provide header class of the DLT protocol. """
import struct
from enum import IntEnum
from typing import Optional, cast

###############################################################################
# Standard Header of the DLT protocol
# It can be checked at following sections:
# - 7.7.2 Header Definition of the Dlt Protocol
# - 7.7.3 Standard Header
# in AUTOSAR Specification of Diagnostic Log and Trace V1.2.0 R4.0 Rev3
###############################################################################


class StandardHeader:
    """The Standard Header of a DLT Message."""

    # length of the bytes data
    DATA_MIN_LENGTH = 4

    # bit masks to get/set values in Header Type
    USE_EXTENDED_HEADER_MASK = 0b00000001
    MSB_FIRST_MASK = 0b00000010
    WITH_ECU_ID_MASK = 0b00000100
    WITH_SESSION_ID_MASK = 0b00001000
    WITH_TIMESTAMP_MASK = 0b00010000
    VERSION_NUMBER_MASK = 0b11100000

    _VERSION_NUMBER_SHIFT = 5

    # struct format for pack/unpack
    STRUCT_MIN_FORMAT = ">BBH"

    def __init__(
        self,
        use_extended_header: bool,
        msb_first: bool,
        version_number: int,
        message_counter: int,
        length: int,
        ecu_id: Optional[str] = None,
        session_id: Optional[int] = None,
        timestamp: Optional[int] = None,
    ) -> None:
        """Create StandardHeader object.

        Args:
            use_extended_header (bool): If set, the Extended Header is transmitted.
            msb_first (bool): If set, the payload data is in big endian format,
                              else in little endian format.
            version_number (int): Version number of Dlt Data protocol
            message_counter (int): Continuous number of message
            length (int): Length of the complete message in bytes
            ecu_id (Optional[str]): Unique address of sender
            session_id (Optional[int]): Session number
            timestamp (Optional[int]): Continuous time / ticks from the ECU
                                       at the moment the message is sent to Dlt
        """
        self.use_extended_header = use_extended_header
        self.msb_first = msb_first
        self.version_number = version_number
        self.message_counter = message_counter
        self.length = length
        self.ecu_id = ecu_id
        self.session_id = session_id
        self.timestamp = timestamp

    def __repr__(self):
        val = (
            f"StandardHeader(use_extended_header={self.use_extended_header}, "
            f"msb_first={self.msb_first}, version_number={self.version_number}, "
            f"message_counter={self.message_counter}, length={self.length}"
        )
        if self.ecu_id is not None:
            val += f', ecu_id="{self.ecu_id}"'
        if self.session_id is not None:
            val += f", session_id={self.session_id}"
        if self.timestamp is not None:
            val += f", timestamp={self.timestamp}"
        val += ")"
        return val

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (
                self.use_extended_header == other.use_extended_header
                and self.msb_first == other.msb_first
                and self.version_number == other.version_number
                and self.message_counter == other.message_counter
                and self.length == other.length
                and self.ecu_id == other.ecu_id
                and self.session_id == other.session_id
                and self.timestamp == other.timestamp
            )
        return False

    @classmethod
    def create_from_bytes(cls, data: bytes) -> "StandardHeader":
        """Create StandardHeader object from data bytes.

        Args:
            data (bytes): Data bytes

        Raises:
            ValueError: It can be caused by invalid data format.

        Returns:
            StandardHeader: New StandardHeader object
        """
        # validate arguments
        data_length = len(data)
        if data_length < cls.DATA_MIN_LENGTH:
            raise ValueError(
                f"Unexpected length of the data: {data_length} / "
                f"Standard Header must be {cls.DATA_MIN_LENGTH} or more"
            )

        # get header type
        ueh = bool(data[0] & cls.USE_EXTENDED_HEADER_MASK)
        msbf = bool(data[0] & cls.MSB_FIRST_MASK)
        weid = bool(data[0] & cls.WITH_ECU_ID_MASK)
        wsid = bool(data[0] & cls.WITH_SESSION_ID_MASK)
        wtms = bool(data[0] & cls.WITH_TIMESTAMP_MASK)
        vers = int((data[0] & cls.VERSION_NUMBER_MASK) >> cls._VERSION_NUMBER_SHIFT)

        # validate data
        unpack_format = cls.STRUCT_MIN_FORMAT
        expected_data_length = cls.DATA_MIN_LENGTH
        if weid:
            expected_data_length += 4
            unpack_format += "4s"
        if wsid:
            expected_data_length += 4
            unpack_format += "I"
        if wtms:
            expected_data_length += 4
            unpack_format += "I"
        if data_length < expected_data_length:
            raise ValueError(
                f"Unexpected length of the data: {data_length} / "
                f"Standard Header with Header Type: "
                f"WEID={weid} WSID={wsid} WTMS={wtms} "
                f"must be {expected_data_length} or more"
            )

        # parse bytes
        entries = struct.unpack_from(unpack_format, data)
        mcnt = entries[1]
        length = entries[2]
        entry_index = 3
        ecu = None
        if weid:
            ecu = _ascii_decode(entries[entry_index])
            entry_index += 1
        seid = None
        if wsid:
            seid = entries[entry_index]
            entry_index += 1
        tmsp = None
        if wtms:
            tmsp = entries[entry_index]

        return cls(ueh, msbf, vers, mcnt, length, ecu, seid, tmsp)

    def to_bytes(self) -> bytes:
        """Convert to data bytes.

        Returns:
            bytes: Converted data bytes
        """
        data = struct.pack(
            self.STRUCT_MIN_FORMAT, self.header_type, self.message_counter, self.length
        )
        # with_ecu_id is not used to avoid False Positive error of PyRights
        if self.ecu_id is not None:
            data += struct.pack(">4s", _ascii_encode(self.ecu_id))
        if self.with_session_id:
            data += struct.pack(">I", self.session_id)
        if self.with_timestamp:
            data += struct.pack(">I", self.timestamp)
        return data

    @property
    def bytes_length(self) -> int:
        """Get length of the data bytes.

        Returns:
            int: Length of the data bytes
        """
        length = self.DATA_MIN_LENGTH
        if self.with_ecu_id:
            length += 4
        if self.with_session_id:
            length += 4
        if self.with_timestamp:
            length += 4
        return length

    @property
    def header_type(self) -> int:
        """Get Header Type.

        Returns:
            int: Header Type
        """
        htyp = self.version_number << self._VERSION_NUMBER_SHIFT
        if self.use_extended_header:
            htyp |= self.USE_EXTENDED_HEADER_MASK
        if self.msb_first:
            htyp |= self.MSB_FIRST_MASK
        if self.with_ecu_id:
            htyp |= self.WITH_ECU_ID_MASK
        if self.with_session_id:
            htyp |= self.WITH_SESSION_ID_MASK
        if self.with_timestamp:
            htyp |= self.WITH_TIMESTAMP_MASK
        return htyp

    @property
    def with_ecu_id(self) -> bool:
        """Check WEID (With ECU ID) bit is set in the Header Type.

        Returns:
            bool:
                - True: The Standard Header has ECU ID.
                - False: The Standard Header does not have ECU ID.
        """
        return self.ecu_id is not None

    @property
    def with_session_id(self) -> bool:
        """Check WSID (With Session ID) bit is set in the Header Type.

        Returns:
            bool:
                - True: The Standard Header has Session ID.
                - False: The Standard Header does not have Session ID.
        """
        return self.session_id is not None

    @property
    def with_timestamp(self) -> bool:
        """Check WTMS (With Timestamp) bit is set in the Header Type.

        Returns:
            bool:
                - True: The Standard Header has Timestamp.
                - False: The Standard Header does not have Timestamp.
        """
        return self.timestamp is not None


###############################################################################
# Extended Header of the DLT protocol
# It can be checked at following sections:
# - 7.7.2 Header Definition of the Dlt Protocol
# - 7.7.4 Dlt Extended Header
# in AUTOSAR Specification of Diagnostic Log and Trace V1.2.0 R4.0 Rev3
###############################################################################


class MessageType(IntEnum):
    """A definition of Message Type in DLT Extended Header."""

    DLT_TYPE_LOG = 0x0
    DLT_TYPE_APP_TRACE = 0x1
    DLT_TYPE_NW_TRACE = 0x2
    DLT_TYPE_CONTROL = 0x3


class MessageTypeInfo(IntEnum):
    """A definition of Message Type Info in DLT Extended Header."""


class MessageLogInfo(MessageTypeInfo):
    """A definition of Message Log Info in DLT Extended Header."""

    DLT_LOG_FATAL = 0x1
    DLT_LOG_ERROR = 0x2
    DLT_LOG_WARN = 0x3
    DLT_LOG_INFO = 0x4
    DLT_LOG_DEBUG = 0x5
    DLT_LOG_VERBOSE = 0x6


class MessageTraceInfo(MessageTypeInfo):
    """A definition of Message Trace Info in DLT Extended Header."""

    DLT_TRACE_VARIABLE = 0x1
    DLT_TRACE_FUNCTON_IN = 0x2
    DLT_TRACE_FUNCTON_OUT = 0x3
    DLT_TRACE_STATE = 0x4
    DLT_TRACE_VFB = 0x5


class MessageBusInfo(MessageTypeInfo):
    """A definition of Message Trace Info in DLT Extended Header."""

    DLT_NW_TRACE_IPC = 0x1
    DLT_NW_TRACE_CAN = 0x2
    DLT_NW_TRACE_FLEXRAY = 0x3
    DLT_NW_TRACE_MOST = 0x4


class MessageControlInfo(MessageTypeInfo):
    """A definition of Message Control Info in DLT Extended Header."""

    DLT_CONTROL_REQUEST = 0x1
    DLT_CONTROL_RESPONSE = 0x2
    DLT_CONTROL_TIME = 0x3


class ExtendedHeader:
    """The Extended Header of a DLT Message."""

    # length of the bytes data
    DATA_LENGTH = 10

    # bit masks to get/set values in Message Info
    VERBOSE_MASK = 0b00000001
    MESSAGE_TYPE_MASK = 0b00001110
    MESSAGE_TYPE_INFO_MASK = 0b11110000

    _MESSAGE_TYPE_SHIFT = 1
    _MESSAGE_TYPE_INFO_SHIFT = 4

    # struct format for pack/unpack
    STRUCT_FORMAT = ">BB4s4s"

    def __init__(
        self,
        verbose: bool,
        message_type: MessageType,
        message_type_info: MessageTypeInfo,
        number_of_arguments: int,
        application_id: str,
        context_id: str,
    ) -> None:
        """Create ExtendedHeader object.

        Args:
            verbose (bool): If set, a description of the transmitted data is provided
                            within the payload.
                            If not set, this information will be given within a file.
            message_type (MessageType): A field describes the transmitted Dlt message
            message_type_info (MessageTypeInfo): A field depends on the Message Type
            number_of_arguments (int): Number of arguments in the message payload
            application_id (str): Number / ID of application
            context_id (str): Unique ID of logging / tracing context
        """
        self.verbose = verbose
        self.message_type = message_type
        self.message_type_info = message_type_info
        self.number_of_arguments = number_of_arguments
        self.application_id = application_id
        self.context_id = context_id

    def __repr__(self):
        return (
            f"ExtendedHeader(verbose={self.verbose}, "
            f"message_type={self.message_type}, "
            f"message_type_info={self.message_type_info}, "
            f"number_of_arguments={self.number_of_arguments}, "
            f'application_id="{self.application_id}", '
            f'context_id="{self.context_id}")'
        )

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (
                self.verbose == other.verbose
                and self.message_type == other.message_type
                and self.message_type_info == other.message_type_info
                and self.number_of_arguments == other.number_of_arguments
                and self.application_id == other.application_id
                and self.context_id == other.context_id
            )
        return False

    @classmethod
    def create_from_bytes(cls, data: bytes) -> "ExtendedHeader":
        """Create ExtendedHeader object from data bytes.

        Args:
            data (bytes): Data bytes

        Raises:
            ValueError: It can be caused by invalid data format.

        Returns:
            ExtendedHeader: New ExtendedHeader object
        """
        # validate arguments
        data_length = len(data)
        if data_length < cls.DATA_LENGTH:
            raise ValueError(
                f"Unexpected length of the data: {data_length} / "
                f"Extended Header must be {cls.DATA_LENGTH} or more"
            )

        # parse bytes
        entries = struct.unpack_from(cls.STRUCT_FORMAT, data)
        msin = entries[0]
        verb = bool(msin & cls.VERBOSE_MASK)
        mstp = (msin & cls.MESSAGE_TYPE_MASK) >> cls._MESSAGE_TYPE_SHIFT
        mtin = (msin & cls.MESSAGE_TYPE_INFO_MASK) >> cls._MESSAGE_TYPE_INFO_SHIFT
        noar = entries[1]
        apid = _ascii_decode(entries[2])
        ctid = _ascii_decode(entries[3])

        return cls(verb, mstp, mtin, noar, apid, ctid)

    def to_bytes(self) -> bytes:
        """Convert to data bytes.

        Returns:
            bytes: Converted data bytes
        """
        msin = 0
        if self.verbose:
            msin |= self.VERBOSE_MASK
        msin |= (self.message_type << self._MESSAGE_TYPE_SHIFT) & self.MESSAGE_TYPE_MASK
        msin |= (
            self.message_type_info << self._MESSAGE_TYPE_INFO_SHIFT
        ) & self.MESSAGE_TYPE_INFO_MASK
        return struct.pack(
            self.STRUCT_FORMAT,
            msin,
            self.number_of_arguments,
            _ascii_encode(self.application_id),
            _ascii_encode(self.context_id),
        )

    @property
    def bytes_length(self) -> int:
        """Get length of the data bytes.

        Returns:
            int: Length of the data bytes
        """
        return self.DATA_LENGTH

    @property
    def message_log_info(self) -> MessageLogInfo:
        """Get Message Log Info.

        message_type_info is converted to MessageLogInfo.

        Returns:
            MessageLogInfo: Message Log Info
        """
        return cast(MessageLogInfo, self.message_type_info)

    @property
    def message_trace_info(self) -> MessageTraceInfo:
        """Get Message Trace Info.

        message_type_info is converted to MessageTraceInfo.

        Returns:
            MessageTraceInfo: Message Trace Info
        """
        return cast(MessageTraceInfo, self.message_type_info)

    @property
    def message_bus_info(self) -> MessageBusInfo:
        """Get Message Bus Info.

        message_type_info is converted to MessageBusInfo.

        Returns:
            MessageBusInfo: Message Bus Info
        """
        return cast(MessageBusInfo, self.message_type_info)

    @property
    def message_control_info(self) -> MessageControlInfo:
        """Get Message Control Info.

        message_type_info is converted to MessageControlInfo.

        Returns:
            MessageControlInfo: Message Control Info
        """
        return cast(MessageControlInfo, self.message_type_info)


###############################################################################
# Storage Header of the DLT protocol
# It can be checked at following sections:
# - 7.7.6 Additional Message Parts
# in AUTOSAR Specification of Diagnostic Log and Trace V1.2.0 R4.0 Rev3
###############################################################################


class StorageHeader:
    """The Storage Header of a DLT Message."""

    # length of the bytes data
    DATA_LENGTH = 16

    # "DLT"+0x01
    DLT_PATTERN = b"\x44\x4C\x54\x01"

    # struct format for pack/unpack
    STRUCT_FORMAT = "<Ii4s"

    def __init__(self, seconds: int, microseconds: int, ecu_id: str) -> None:
        """Create StorageHeader object.

        Args:
            seconds (int): A seconds since 01.01.1970 (unix time)
            microseconds (int): A microseconds of the second (between 0 – 1.000.000)
            ecu_id (str): The ECU ID
        """
        self.seconds = seconds
        self.microseconds = microseconds
        self.ecu_id = ecu_id

    def __repr__(self):
        return (
            f"StorageHeader(seconds={self.seconds}, "
            f'microseconds={self.microseconds}, ecu_id="{self.ecu_id}")'
        )

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (
                self.seconds == other.seconds
                and self.microseconds == other.microseconds
                and self.ecu_id == other.ecu_id
            )
        return False

    @classmethod
    def create_from_bytes(cls, data: bytes) -> "StorageHeader":
        """Create StorageHeader object from data bytes.

        Args:
            data (bytes): Data bytes

        Raises:
            ValueError: It can be caused by invalid data format.

        Returns:
            StorageHeader: New StorageHeader object
        """
        # validate arguments
        data_length = len(data)
        if data_length < cls.DATA_LENGTH:
            raise ValueError(
                f"Unexpected length of the data: {data_length} / "
                f"Storage Header must be {cls.DATA_LENGTH} or more"
            )

        # parse bytes
        dlt_pattern = data[:4]
        if dlt_pattern != cls.DLT_PATTERN:
            raise ValueError(
                f"DLT-Pattern is not found in the data: {dlt_pattern} / "
                f"Beginning of Storage Header must be {cls.DLT_PATTERN}"
            )
        entries = struct.unpack_from(cls.STRUCT_FORMAT, data, 4)
        seconds = entries[0]
        microseconds = entries[1]
        ecu_id = _ascii_decode(entries[2])

        return cls(seconds, microseconds, ecu_id)

    def to_bytes(self) -> bytes:
        """Convert to data bytes.

        Returns:
            bytes: Converted data bytes
        """
        return self.DLT_PATTERN + struct.pack(
            self.STRUCT_FORMAT,
            self.seconds,
            self.microseconds,
            _ascii_encode(self.ecu_id),
        )

    @property
    def bytes_length(self) -> int:
        """Get length of the data bytes.

        Returns:
            int: Length of the data bytes
        """
        return self.DATA_LENGTH


def _ascii_decode(ascii: bytes) -> str:
    """Decode bytes of ASCII charactors to string.

    Args:
        ascii (bytes): ASCII charactors

    Returns:
        str: Converted string
    """
    return ascii.decode("ascii", "replace").replace("\x00", "")


def _ascii_encode(ascii: str) -> bytes:
    """Encode string of ASCII charactors to bytes.

    Args:
        ascii (str): ASCII charactors

    Returns:
        str: Converted bytes
    """
    return ascii.encode("ascii", "replace")



# -*- coding: utf-8 -*-
"""
@Author : zhang.yonggang
@File   : setup.py.py
@Project: pydlt_plus
@Time   : 2023-10-26 08:28:01
@Desc   : The file is ...
@Version: v1.0
"""
from setuptools import setup

setup(
    name='pydlt_plus',
    version='1.1.0',
    author='Arthur Nostmabole Zhang',
    author_email='nostmabole@sina.com',
    description='A class to read and filter DLT message from DLT file BASE ON DltFileReader.',
    long_description='''A class to read and filter DLT message from DLT file BASE ON DltFileReader.''',
    url='https://gitee.com/nostmabole/pydlt_plus',
    packages=['pydlt_plus'],
    install_requires=[
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)



""" Provide message class of the DLT protocol. """
from datetime import datetime, timezone
from typing import List, Optional, cast

from pydlt_plus.header import (
    ExtendedHeader,
    MessageBusInfo,
    MessageControlInfo,
    MessageLogInfo,
    MessageTraceInfo,
    MessageType,
    MessageTypeInfo,
    StandardHeader,
    StorageHeader,
)
from pydlt_plus.payload import Argument, NonVerbosePayload, Payload, VerbosePayload

###############################################################################
# DLT Message of the DLT protocol
# It can be checked at following sections:
# - 7.7.1 Dlt Message Format in General
# - 7.7.6 Additional Message Parts
# in AUTOSAR Specification of Diagnostic Log and Trace V1.2.0 R4.0 Rev3
###############################################################################


_MESSAGE_TYPE_STR = {
    MessageType.DLT_TYPE_LOG: "log",
    MessageType.DLT_TYPE_APP_TRACE: "app_trace",
    MessageType.DLT_TYPE_NW_TRACE: "nw_trace",
    MessageType.DLT_TYPE_CONTROL: "control",
}

_MESSAGE_LOG_INFO_STR = {
    MessageLogInfo.DLT_LOG_FATAL: "fatal",
    MessageLogInfo.DLT_LOG_ERROR: "error",
    MessageLogInfo.DLT_LOG_WARN: "warn",
    MessageLogInfo.DLT_LOG_INFO: "info",
    MessageLogInfo.DLT_LOG_DEBUG: "debug",
    MessageLogInfo.DLT_LOG_VERBOSE: "verbose",
}

_MESSAGE_TRACE_INFO_STR = {
    MessageTraceInfo.DLT_TRACE_VARIABLE: "variable",
    MessageTraceInfo.DLT_TRACE_FUNCTON_IN: "func_in",
    MessageTraceInfo.DLT_TRACE_FUNCTON_OUT: "func_out",
    MessageTraceInfo.DLT_TRACE_STATE: "state",
    MessageTraceInfo.DLT_TRACE_VFB: "vfb",
}

_MESSAGE_BUS_INFO_STR = {
    MessageBusInfo.DLT_NW_TRACE_IPC: "ipc",
    MessageBusInfo.DLT_NW_TRACE_CAN: "can",
    MessageBusInfo.DLT_NW_TRACE_FLEXRAY: "flexray",
    MessageBusInfo.DLT_NW_TRACE_MOST: "most",
}

_MESSAGE_CONTROL_INFO_STR = {
    MessageControlInfo.DLT_CONTROL_REQUEST: "request",
    MessageControlInfo.DLT_CONTROL_RESPONSE: "response",
    MessageControlInfo.DLT_CONTROL_TIME: "time",
}


class DltMessage:
    """A class to handle DLT Message."""

    def __init__(
        self,
        str_header: Optional[StorageHeader],
        std_header: StandardHeader,
        ext_header: Optional[ExtendedHeader],
        payload: Optional[Payload],
    ) -> None:
        """Create DltMessage object.

        In most cases, the constructor do not have to be called directly.
        The following factory methods can be used instead of it:
        - create_non_verbose_message
        - create_verbose_message
        - create_from_bytes

        Args:
            str_header (Optional[StorageHeader]): Storage Header of the message
            std_header (StandardHeader): Standard Header of the message
            ext_header (Optional[ExtendedHeader]): Extended Header of the message
            payload (Optional[Payload]): Payload of the message
        """
        self.str_header = str_header
        self.std_header = std_header
        self.ext_header = ext_header
        self.payload = payload

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (
                self.str_header == other.str_header
                and self.std_header == other.std_header
                and self.ext_header == other.ext_header
                and self.payload == other.payload
            )
        return False

    def __str__(self) -> str:
        """Get overview of the message.

        Overview is string of the following format:
        [<Time> ][<Timestamp> ]<Count>[ <Ecuid>][ <Apid>][ <Ctid>][ <SessionId>]
        [ <Type>][ <Subtype>][ <Mode>][ <#Args>][ <Payload>]

        Returns:
            str: Overview of the message
        """
        ret = []
        if self.str_header is not None:
            ret.append(
                datetime.fromtimestamp(self.str_header.seconds, timezone.utc).strftime(
                    "%Y/%m/%d %H:%M:%S"
                )
                + "."
                + str(self.str_header.microseconds).zfill(6)
            )
        if self.std_header.timestamp is not None:
            ret.append(
                str(int(self.std_header.timestamp / 10000))
                + "."
                + str(int(self.std_header.timestamp % 10000)).zfill(4)
            )
        ret.append(str(self.std_header.message_counter))
        if self.std_header.ecu_id is not None:
            ret.append(self.std_header.ecu_id)
        elif self.str_header is not None:
            ret.append(self.str_header.ecu_id)
        if self.ext_header is not None:
            ret.append(self.ext_header.application_id)
            ret.append(self.ext_header.context_id)
        if self.std_header.session_id is not None:
            ret.append(str(self.std_header.session_id))
        if self.ext_header is not None:
            ret.append(_MESSAGE_TYPE_STR.get(self.ext_header.message_type, "unknown"))
            if self.ext_header.message_type == MessageType.DLT_TYPE_LOG:
                ret.append(
                    _MESSAGE_LOG_INFO_STR.get(
                        self.ext_header.message_log_info, "unknown"
                    )
                )
            if self.ext_header.message_type == MessageType.DLT_TYPE_APP_TRACE:
                ret.append(
                    _MESSAGE_TRACE_INFO_STR.get(
                        self.ext_header.message_trace_info, "unknown"
                    )
                )
            if self.ext_header.message_type == MessageType.DLT_TYPE_NW_TRACE:
                ret.append(
                    _MESSAGE_BUS_INFO_STR.get(
                        self.ext_header.message_bus_info, "unknown"
                    )
                )
            if self.ext_header.message_type == MessageType.DLT_TYPE_CONTROL:
                ret.append(
                    _MESSAGE_CONTROL_INFO_STR.get(
                        self.ext_header.message_control_info, "unknown"
                    )
                )
        if self.ext_header is not None and self.ext_header.verbose:
            ret.append("verbose")
        else:
            ret.append("non-verbose")
        if self.ext_header is not None:
            ret.append(str(self.ext_header.number_of_arguments))
        if self.payload is not None:
            ret.append(str(self.payload))
        return " ".join(ret)

    @classmethod
    def create_non_verbose_message(
        cls,
        message_id: int,
        non_static_data: bytes,
        ext_header: Optional[ExtendedHeader] = None,
        timestamp: Optional[int] = None,
        session_id: Optional[int] = None,
        ecu_id: Optional[str] = None,
        message_counter: int = 0,
        version_number: int = 1,
        msb_first: bool = False,
        str_header: StorageHeader = None,
    ) -> "DltMessage":
        """Create DltMessage object as Non-Verbose mode.

        Returns:
            DltMessage: DltMessage object as Non-Verbose mode
        """
        payload = NonVerbosePayload(message_id, non_static_data, msb_first)
        std_header = cls._create_standard_header(
            payload.bytes_length,
            ext_header,
            timestamp,
            session_id,
            ecu_id,
            message_counter,
            version_number,
            msb_first,
        )
        return cls(str_header, std_header, ext_header, payload)

    @classmethod
    def create_verbose_message(
        cls,
        arguments: List[Argument],
        message_type: MessageType,
        message_type_info: MessageTypeInfo,
        application_id: str,
        context_id: str,
        timestamp: Optional[int] = None,
        session_id: Optional[int] = None,
        ecu_id: Optional[str] = None,
        message_counter: int = 0,
        version_number: int = 1,
        msb_first: bool = False,
        str_header: StorageHeader = None,
    ) -> "DltMessage":
        """Create DltMessage object as Verbose mode.

        Returns:
            DltMessage: DltMessage object as Verbose mode
        """
        payload = VerbosePayload(arguments)
        ext_header = ExtendedHeader(
            True,
            message_type,
            message_type_info,
            len(arguments),
            application_id,
            context_id,
        )
        std_header = cls._create_standard_header(
            payload.bytes_length,
            ext_header,
            timestamp,
            session_id,
            ecu_id,
            message_counter,
            version_number,
            msb_first,
        )
        return cls(str_header, std_header, ext_header, payload)

    @staticmethod
    def _create_standard_header(
        payload_length: int,
        ext_header: Optional[ExtendedHeader],
        timestamp: Optional[int],
        session_id: Optional[int],
        ecu_id: Optional[str],
        message_counter: int,
        version_number: int,
        msb_first: bool,
    ) -> StandardHeader:
        if ext_header is None:
            use_extended_header = False
            ext_header_length = 0
        else:
            use_extended_header = True
            ext_header_length = ext_header.bytes_length
        std_header = StandardHeader(
            use_extended_header,
            msb_first,
            version_number,
            message_counter,
            0,  # temporary length; it will be set on the next statement
            ecu_id,
            session_id,
            timestamp,
        )
        std_header.length = payload_length + ext_header_length + std_header.bytes_length
        return std_header

    @classmethod
    def create_from_bytes(
        cls, data: bytes, with_storage_header: bool, encoding: Optional[str] = None
    ) -> "DltMessage":
        """Create DltMessage object from data bytes.

        Args:
            data (bytes): Data bytes
            with_storage_header (bool): The data has storage header or not

        Raises:
            ValueError: It can be caused by invalid data format.

        Returns:
            DltMessage: New DltMessage object
        """
        seek_pos = 0
        str_header = None
        str_header_length = 0
        if with_storage_header:
            str_header = StorageHeader.create_from_bytes(data)
            seek_pos += str_header.bytes_length
            str_header_length = str_header.bytes_length
        std_header = StandardHeader.create_from_bytes(data[seek_pos:])
        seek_pos += std_header.bytes_length
        ext_header = None
        ext_header_length = 0
        if std_header.use_extended_header:
            ext_header = ExtendedHeader.create_from_bytes(data[seek_pos:])
            ext_header_length = ext_header.bytes_length
            seek_pos += ext_header.bytes_length
        payload = None
        if std_header.length > std_header.bytes_length + ext_header_length:
            is_verbose = ext_header is not None and ext_header.verbose is True
            if is_verbose:
                payload = VerbosePayload.create_from_bytes(
                    data[seek_pos : std_header.length + str_header_length],
                    std_header.msb_first,
                    ext_header.number_of_arguments,
                    encoding,
                )
            else:
                payload = NonVerbosePayload.create_from_bytes(
                    data[seek_pos : std_header.length + str_header_length],
                    std_header.msb_first,
                )
        return cls(str_header, std_header, ext_header, payload)

    def to_bytes(self) -> bytes:
        """Convert to data bytes.

        Returns:
            bytes: Converted data bytes
        """
        data = b""
        if self.str_header is not None:
            data += self.str_header.to_bytes()
        data += self.std_header.to_bytes()
        if self.ext_header is not None:
            data += self.ext_header.to_bytes()
        if self.payload is not None:
            data += self.payload.to_bytes(self.std_header.msb_first)
        return data

    @property
    def verbose(self) -> bool:
        """Check the message is verbose mode or non-verbose mode.

        Returns:
            bool:
                - False: Non-Verbose mode
                -  True: Verbose mode
        """
        return False if self.ext_header is None else self.ext_header.verbose

    @property
    def non_verbose_payload(self) -> NonVerbosePayload:
        """Get payload as non-verbose mode.

        Unexpected behaviors can be caused if the DltMessage is verbose mode.

        Returns:
            NonVerbosePayload: Payload as non-verbose mode
        """
        return cast(NonVerbosePayload, self.payload)

    @property
    def verbose_payload(self) -> VerbosePayload:
        """Get payload as verbose mode.

        Unexpected behaviors can be caused if the DltMessage is non-verbose mode.

        Returns:
            NonVerbosePayload: Payload as verbose mode
        """
        return cast(VerbosePayload, self.payload)



""" Provide class to handle DLT file. """
import struct
from pathlib import Path
from typing import Iterator, List, Optional, Union

from pydlt_plus.header import StandardHeader, StorageHeader
from pydlt_plus.message import DltMessage


class DltFileReader:
    """A class to read DLT message from DLT file.

    Examples::
        # create reader as context manager
        with DltFileReader("filepath") as reader:
            # read 1 message from file
            message = reader.read_message()

            # read all messages from file
            messages = reader.read_messages()

        # create reader as iterator
        for message in DltFileReader("filepath"):  # read all messages
            # handle each message
    """

    def __init__(self, path: Union[str, Path], encoding: Optional[str] = None) -> None:
        """Create DltFileReader object.

        Open a file of the path in the constructor.
        The file should be closed by calling close() method
        when the file is no longer used by the class.

        The class supports context manager and with block.
        close() method does not have to be called if using it.

        Args:
            path (Union[str, Path]): A path to file.
            encoding: encoding that will be used for parsing non-utf-8 dlt strings
                      The dlt specification only supports ascii and utf-8 explicitly.
                      However, some implementations store dlt strings in a local 8-bit
                      format (e.g. latin-1) instead of plain ascii.
        """
        self._file = open(str(path), "rb")
        self._encoding = encoding

    def __enter__(self) -> "DltFileReader":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> Optional[bool]:
        self.close()

    def close(self) -> None:
        """Close a file opened by the class."""
        self._file.close()

    @property
    def closed(self) -> bool:
        """Check a file opened by the class is closed.

        Returns:
            bool: A file is closed if True.
        """
        return self._file.closed

    def __iter__(self) -> Iterator[DltMessage]:
        return self

    def __next__(self) -> DltMessage:
        message = self.read_message()
        if message is None:
            raise StopIteration()
        return message

    def read_message(self) -> Optional[DltMessage]:
        """Read 1 DLT message from file.

        Returns:
            Optional[DltMessage]: DLT message or None if not enough data to read
        """
        min_length = StorageHeader.DATA_LENGTH + StandardHeader.DATA_MIN_LENGTH
        msg_data = self._file.read(min_length)

        if len(msg_data) < min_length:
            return None
        length = struct.unpack_from(
            StandardHeader.STRUCT_MIN_FORMAT, msg_data, StorageHeader.DATA_LENGTH
        )[2]
        msg_length = StorageHeader.DATA_LENGTH + length
        msg_data += self._file.read(msg_length - min_length)
        if len(msg_data) < msg_length:
            return None
        return DltMessage.create_from_bytes(msg_data, True, self._encoding)

    def read_messages(self) -> List[DltMessage]:
        """Read all DLT messages from file.

        Returns:
            List[DltMessage]: All DLT messages in the file
        """
        return [message for message in self.__iter__()]


class DltFileWriter:
    """A class to write DLT message to DLT file.

    Examples::
        # create writer as context manager
        with DltFileWriter("filepath") as writer:
            # write 1 message to file
            writer.write_message(message)

            # write messages to file
            writer.write_messages(messages)
    """

    def __init__(self, path: Union[str, Path], append: bool = False) -> None:
        """Create DltFileWriter object.

        Open a file of the path in the constructor.
        The file should be closed by calling close() method
        when the file is no longer used by the class.

        The class supports context manager and with block.
        close() method does not have to be called if using it.

        Args:
            path (Union[str, Path]): A path to file.
            append (bool, optional): Set True if append mode. Defaults to False.
        """
        mode = "ab" if append else "wb"
        self._file = open(path, mode)

    def __enter__(self) -> "DltFileWriter":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> Optional[bool]:
        self.close()

    def close(self) -> None:
        """Close a file opened by the class."""
        self._file.close()

    @property
    def closed(self) -> bool:
        """Check a file opened by the class is closed.

        Returns:
            bool: A file is closed if True.
        """
        return self._file.closed

    def write_message(self, message: DltMessage) -> None:
        """Write 1 DLT message to file.

        Args:
            message (DltMessage): DLT message
        """
        self._file.write(message.to_bytes())

    def write_messages(self, messages: List[DltMessage]) -> None:
        """Write DLT messages to file.

        Args:
            messages (List[DltMessage]): DLT messages
        """
        [self.write_message(message) for message in messages]



"""Provide payload class of the DLT protocol. """
import struct
from abc import ABC, abstractmethod
from enum import IntEnum
from typing import List, Optional, Union

###############################################################################
# Payload of the DLT protocol
# It can be checked at following sections:
# - 7.7.5 Payload
# in AUTOSAR Specification of Diagnostic Log and Trace V1.2.0 R4.0 Rev3
###############################################################################

msb_first_default: Optional[bool] = False
"""Default value of endian of the payload data

The value will be used when casting the object to bytes,
if endian is not set by argument and field value of endian type is not defined.
"""


class Payload(ABC):
    """The Payload of a DLT Message."""

    def __str__(self) -> str:
        return self._to_str()

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.to_bytes(False) == other.to_bytes(False)
        return False

    @abstractmethod
    def to_bytes(self, msb_first: Optional[bool] = None) -> bytes:
        """Convert to data bytes.

        Args:
            msb_first:  True - Handle payload data as big endian
                       False - Handle payload data as little endian
                        None - Handle payload data according to field value of endian
                               type. The value of msb_first_default is used
                               if field value of endian type is also None.
                               Raise ValueError if both are None.

        Raises:
            ValueError: All endian type of the argument, field value,
                        and msb_first_default are None

        Returns:
            bytes: Converted data bytes
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def bytes_length(self) -> int:
        """Get length of the data bytes.

        Returns:
            int: Length of the data bytes
        """
        raise NotImplementedError

    @abstractmethod
    def _to_str(self) -> str:
        """Convert payload to human readable string.

        Returns:
            str: Human readable string.
        """
        raise NotImplementedError


###############################################################################
# Payload: Non-Verbose Mode of the DLT protocol
# It can be checked at following sections:
# - 7.7.5.1 Non-Verbose Mode
# in AUTOSAR Specification of Diagnostic Log and Trace V1.2.0 R4.0 Rev3
###############################################################################


class NonVerbosePayload(Payload):
    """The Payload of Non-Verbose Mode of a DLT Message."""

    # minimum length of the bytes data
    _MESSAGE_ID_LENGTH = 4

    def __init__(
        self, message_id: int, non_static_data: bytes, msb_first: Optional[bool] = None
    ) -> None:
        """Create NonVerbosePayload object.

        Args:
            message_id (int): Unique ID for a specific Dlt message
            non_static_data (bytes): All non-static information of a Dlt message.
            msb_first (Optional[bool]): If set, the payload data is in big endian,
                                        else in little endian.
        """
        self.message_id = message_id
        self.non_static_data = non_static_data
        self.msb_first = msb_first

    @classmethod
    def create_from_bytes(cls, data: bytes, msb_fitst: bool) -> "NonVerbosePayload":
        """Create NonVerbosePayload object from data bytes.

        Args:
            data (bytes): Data bytes
            msb_first (bool): If set, the payload data is in big endian format,
                              else in little endian format.

        Raises:
            ValueError: It can be caused by invalid data format.

        Returns:
            NonVerbosePayload: New NonVerbosePayload object
        """
        # validate arguments
        data_length = len(data)
        if data_length < cls._MESSAGE_ID_LENGTH:
            raise ValueError(
                f"Unexpected length of the data: {data_length} / "
                f"Payload of Non-Verbose Mode must not be < {cls._MESSAGE_ID_LENGTH}"
            )

        # parse bytes
        endian = ">" if msb_fitst else "<"
        entries = struct.unpack(f"{endian}I", data[:4])
        message_id = entries[0]

        return cls(message_id, data[4:], msb_fitst)

    def to_bytes(self, msb_first: Optional[bool] = None) -> bytes:
        """Convert to data bytes.

        Args:
            msb_first:  True - Handle payload data as big endian
                       False - Handle payload data as little endian
                        None - Handle payload data according to field value of endian
                               type. The value of msb_first_default is used
                               if field value of endian type is also None.
                               Raise ValueError if both are None.

        Raises:
            ValueError: All endian type of the argument, field value,
                        and msb_first_default are None

        Returns:
            bytes: Converted data bytes
        """
        if msb_first is not None:
            endian = ">" if msb_first else "<"
        elif self.msb_first is not None:
            endian = ">" if self.msb_first else "<"
        elif msb_first_default is not None:
            endian = ">" if msb_first_default else "<"
        else:
            raise ValueError("Endian is not known")
        return struct.pack(f"{endian}I", self.message_id) + self.non_static_data

    @property
    def bytes_length(self) -> int:
        """Get length of the data bytes.

        Returns:
            int: Length of the data bytes
        """
        return self._MESSAGE_ID_LENGTH + len(self.non_static_data)

    def _to_str(self) -> str:
        """Convert payload to human readable string.

        Returns:
            str: Human readable string.
        """
        return f"[{self.message_id}] {self.non_static_data.hex()}"


###############################################################################
# Payload: Verbose Mode of the DLT protocol
# It can be checked at following sections:
# - 7.7.5.2 Verbose Mode
# in AUTOSAR Specification of Diagnostic Log and Trace V1.2.0 R4.0 Rev3
###############################################################################


# Base type:
# - Type Bool (BOOL)
# - Type Signed (SINT)
# - Type Unsigned (UINT)
# - Type Float (FLOA)
# - Type Array (ARAY)    *not supported*
# - Type String (STRG)
# - Type Raw (RAWD)
# - Trace Info (TRAI)    *not supported*
# - Type Struct (STRU)   *not supported*
# Option type:
# - Type Length (TYLE)   *supported with limitation*
# - Variable Info (VARI) *not supported*
# - Fixed Point (FIXP)   *not supported*
# - String Coding (SCOD)
# fmt: off
MASK_BASE_TYPE     = 0b00000000000000000110011111110000  # noqa: E221
MASK_TYPE_LENGTH   = 0b00000000000000000000000000001111  # noqa: E221
MASK_VARIABLE_INFO = 0b00000000000000000000100000000000  # noqa: E221
MASK_FIXED_POINT   = 0b00000000000000000001000000000000  # noqa: E221
MASK_STRING_CODING = 0b00000000000000111000000000000000  # noqa: E221
# fmt: on


class TypeInfo(IntEnum):
    """A definition of Type Info of the Argument of Verbose Mode Paylaod.

    Note: The following format has not been supported:
    - TYPE_LENGTH_128BIT
    - TYPE_ARRAY
    - VARIABLE_INFO
    - FIXED_POINT
    - TRACE_INFO
    - TYPE_STRUCT
    """

    # fmt: off
    # Type Length
    TYPE_LENGTH_8BIT    = 0b00000000000000000000000000000001    # noqa: E221
    TYPE_LENGTH_16BIT   = 0b00000000000000000000000000000010    # noqa: E221
    TYPE_LENGTH_32BIT   = 0b00000000000000000000000000000011    # noqa: E221
    TYPE_LENGTH_64BIT   = 0b00000000000000000000000000000100    # noqa: E221
    TYPE_LENGTH_128BIT  = 0b00000000000000000000000000000101    # noqa: E221
    # Type Bool
    TYPE_BOOL           = 0b00000000000000000000000000010000    # noqa: E221
    # Type Signed
    TYPE_SIGNED         = 0b00000000000000000000000000100000    # noqa: E221
    # Type Unsigned
    TYPE_UNSIGNED       = 0b00000000000000000000000001000000    # noqa: E221
    # Type Float
    TYPE_FLOAT          = 0b00000000000000000000000010000000    # noqa: E221
    # Type Array
    TYPE_ARRAY          = 0b00000000000000000000000100000000    # noqa: E221
    # Type String
    TYPE_STRING         = 0b00000000000000000000001000000000    # noqa: E221
    # Type Raw
    TYPE_RAW            = 0b00000000000000000000010000000000    # noqa: E221
    # Variable Info
    VARIABLE_INFO       = 0b00000000000000000000100000000000    # noqa: E221
    # Fixed Point
    FIXED_POINT         = 0b00000000000000000001000000000000    # noqa: E221
    # Trace Info
    TRACE_INFO          = 0b00000000000000000010000000000000    # noqa: E221
    # Type Struct
    TYPE_STRUCT         = 0b00000000000000000100000000000000    # noqa: E221
    # String Coding
    STRING_CODING_ASCII = 0b00000000000000000000000000000000    # noqa: E221
    STRING_CODING_UTF8  = 0b00000000000000001000000000000000    # noqa: E221
    # fmt: on


class Argument(ABC):

    _TYPE_INFO_LENGTH = 4

    def __init__(self, msb_first: Optional[bool]):
        self.msb_first = msb_first

    def __str__(self) -> str:
        return self._to_str()

    @property
    @abstractmethod
    def _type_info(self) -> int:
        raise NotImplementedError

    @classmethod
    def create_from_bytes(
        cls, data: bytes, msb_first: bool, encoding: Optional[str] = None
    ) -> "Argument":
        """Create Argument object from data bytes.

        Raises:
            ValueError: Unsupported TypeInfo is in the data.

        Returns:
            Argument: Argument object
        """
        endian = ">" if msb_first else "<"
        type_info = struct.unpack(f"{endian}I", data[: cls._TYPE_INFO_LENGTH])[0]
        type_info_base = type_info & MASK_BASE_TYPE
        if type_info_base == TypeInfo.TYPE_BOOL:
            return ArgumentBool.from_data_payload(
                data[cls._TYPE_INFO_LENGTH :], msb_first
            )
        elif type_info_base == TypeInfo.TYPE_SIGNED:
            type_info_length = type_info & MASK_TYPE_LENGTH
            if type_info_length == TypeInfo.TYPE_LENGTH_8BIT:
                return ArgumentSInt8.from_data_payload(
                    data[cls._TYPE_INFO_LENGTH :], msb_first
                )
            elif type_info_length == TypeInfo.TYPE_LENGTH_16BIT:
                return ArgumentSInt16.from_data_payload(
                    data[cls._TYPE_INFO_LENGTH :], msb_first
                )
            elif type_info_length == TypeInfo.TYPE_LENGTH_32BIT:
                return ArgumentSInt32.from_data_payload(
                    data[cls._TYPE_INFO_LENGTH :], msb_first
                )
            elif type_info_length == TypeInfo.TYPE_LENGTH_64BIT:
                return ArgumentSInt64.from_data_payload(
                    data[cls._TYPE_INFO_LENGTH :], msb_first
                )
        elif type_info_base == TypeInfo.TYPE_UNSIGNED:
            type_info_length = type_info & MASK_TYPE_LENGTH
            if type_info_length == TypeInfo.TYPE_LENGTH_8BIT:
                return ArgumentUInt8.from_data_payload(
                    data[cls._TYPE_INFO_LENGTH :], msb_first
                )
            elif type_info_length == TypeInfo.TYPE_LENGTH_16BIT:
                return ArgumentUInt16.from_data_payload(
                    data[cls._TYPE_INFO_LENGTH :], msb_first
                )
            elif type_info_length == TypeInfo.TYPE_LENGTH_32BIT:
                return ArgumentUInt32.from_data_payload(
                    data[cls._TYPE_INFO_LENGTH :], msb_first
                )
            elif type_info_length == TypeInfo.TYPE_LENGTH_64BIT:
                return ArgumentUInt64.from_data_payload(
                    data[cls._TYPE_INFO_LENGTH :], msb_first
                )
        elif type_info_base == TypeInfo.TYPE_FLOAT:
            type_info_length = type_info & MASK_TYPE_LENGTH
            if type_info_length == TypeInfo.TYPE_LENGTH_32BIT:
                return ArgumentFloat32.from_data_payload(
                    data[cls._TYPE_INFO_LENGTH :], msb_first
                )
            elif type_info_length == TypeInfo.TYPE_LENGTH_64BIT:
                return ArgumentFloat64.from_data_payload(
                    data[cls._TYPE_INFO_LENGTH :], msb_first
                )
        elif type_info_base == TypeInfo.TYPE_STRING:
            type_info_string_coding = type_info & MASK_STRING_CODING
            if type_info_string_coding == TypeInfo.STRING_CODING_ASCII:
                return ArgumentString.from_data_payload(
                    data[cls._TYPE_INFO_LENGTH :], False, msb_first, encoding
                )
            elif type_info_string_coding == TypeInfo.STRING_CODING_UTF8:
                return ArgumentString.from_data_payload(
                    data[cls._TYPE_INFO_LENGTH :], True, msb_first
                )
        elif type_info_base == TypeInfo.TYPE_RAW:
            return ArgumentRaw.from_data_payload(
                data[cls._TYPE_INFO_LENGTH :], msb_first
            )
        raise ValueError(f"Unsupported TypeInfo: {bin(type_info)}")

    @classmethod
    @abstractmethod
    def from_data_payload(cls, data_payload: bytes, msb_first: bool) -> "Argument":
        """Create Argument object from data payload bytes.

        Args:
            data_payload (bytes): Data payload bytes (without type info)
            msb_first (bool): If set, the payload data is in big endian format,
                              else in little endian format.

        Raises:
            ValueError: It can be caused by invalid data format.

        Returns:
            Argument: New Argument object
        """
        raise NotImplementedError

    def to_bytes(self, msb_first: Optional[bool] = None) -> bytes:
        """Convert to data bytes.

        Args:
            msb_first:  True - Handle payload data as big endian
                       False - Handle payload data as little endian
                        None - Handle payload data according to field value of endian
                               type. The value of msb_first_default is used
                               if field value of endian type is also None.
                               Raise ValueError if both are None.

        Raises:
            ValueError: All endian type of the argument, field value,
                        and msb_first_default are None

        Returns:
            bytes: Converted data bytes
        """
        if msb_first is not None:
            endian = ">" if msb_first else "<"
        elif self.msb_first is not None:
            endian = ">" if self.msb_first else "<"
        elif msb_first_default is not None:
            endian = ">" if msb_first_default else "<"
        else:
            raise ValueError("Endian is not known")
        return struct.pack(f"{endian}I", self._type_info) + self.data_payload_to_bytes(
            msb_first
        )

    @property
    def bytes_length(self) -> int:
        """Get length of the data bytes (including type info).

        Returns:
            int: Length of the data bytes
        """
        return self._TYPE_INFO_LENGTH + self.data_payload_length

    @abstractmethod
    def _to_str(self) -> str:
        """Convert payload to human readable string.

        Returns:
            str: Human readable string.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def data_payload_length(self) -> int:
        """Get length of the data payload.

        Returns:
            int: Length of the data payload
        """
        raise NotImplementedError

    @abstractmethod
    def data_payload_to_bytes(self, msb_first: Optional[bool] = None) -> bytes:
        """Convert data payload to bytes.

        Returns:
            bytes: Convert data payload
        """
        raise NotImplementedError


class ArgumentNumBase(Argument):
    """It is a class for number base argument.

    It has a data field for value.
    A length of the data field is changed by type info.

    """

    def __init__(
        self,
        data: Union[bool, int, float],
        msb_first: Optional[bool] = None,
    ):
        super().__init__(msb_first)
        self.data = data

    @staticmethod
    @abstractmethod
    def _struct_format() -> str:
        """Get format string for struct.pack() / struct.unpack().

        Returns:
            str: Format string for struct.pack() / struct.unpack().

        """
        raise NotImplementedError

    def _to_str(self) -> str:
        return str(self.data)

    @staticmethod
    @abstractmethod
    def _data_payload_length() -> int:
        raise NotImplementedError

    @classmethod
    def from_data_payload(cls, data_payload: bytes, msb_first: bool) -> "Argument":
        endian = ">" if msb_first else "<"
        return cls(
            struct.unpack(
                f"{endian}{cls._struct_format()}",
                data_payload[: cls._data_payload_length()],
            )[0],
            msb_first,
        )

    def data_payload_to_bytes(self, msb_first: Optional[bool] = None) -> bytes:
        if msb_first is not None:
            endian = ">" if msb_first else "<"
        elif self.msb_first is not None:
            endian = ">" if self.msb_first else "<"
        elif msb_first_default is not None:
            endian = ">" if msb_first_default else "<"
        else:
            raise ValueError("Endian is not known")
        return struct.pack(f"{endian}{self._struct_format()}", self.data)


class ArgumentBool(ArgumentNumBase):
    @property
    def _type_info(self) -> int:
        return TypeInfo.TYPE_BOOL | TypeInfo.TYPE_LENGTH_8BIT

    @staticmethod
    def _struct_format() -> str:
        return "?"

    @staticmethod
    def _data_payload_length() -> int:
        return 1

    @property
    def data_payload_length(self) -> int:
        return self._data_payload_length()


class ArgumentUInt8(ArgumentNumBase):
    @property
    def _type_info(self) -> int:
        return TypeInfo.TYPE_UNSIGNED | TypeInfo.TYPE_LENGTH_8BIT

    @staticmethod
    def _struct_format() -> str:
        return "B"

    @staticmethod
    def _data_payload_length() -> int:
        return 1

    @property
    def data_payload_length(self) -> int:
        return self._data_payload_length()


class ArgumentUInt16(ArgumentNumBase):
    @property
    def _type_info(self) -> int:
        return TypeInfo.TYPE_UNSIGNED | TypeInfo.TYPE_LENGTH_16BIT

    @staticmethod
    def _struct_format() -> str:
        return "H"

    @staticmethod
    def _data_payload_length() -> int:
        return 2

    @property
    def data_payload_length(self) -> int:
        return self._data_payload_length()


class ArgumentUInt32(ArgumentNumBase):
    @property
    def _type_info(self) -> int:
        return TypeInfo.TYPE_UNSIGNED | TypeInfo.TYPE_LENGTH_32BIT

    @staticmethod
    def _struct_format() -> str:
        return "I"

    @staticmethod
    def _data_payload_length() -> int:
        return 4

    @property
    def data_payload_length(self) -> int:
        return self._data_payload_length()


class ArgumentUInt64(ArgumentNumBase):
    @property
    def _type_info(self) -> int:
        return TypeInfo.TYPE_UNSIGNED | TypeInfo.TYPE_LENGTH_64BIT

    @staticmethod
    def _struct_format() -> str:
        return "Q"

    @staticmethod
    def _data_payload_length() -> int:
        return 8

    @property
    def data_payload_length(self) -> int:
        return self._data_payload_length()


class ArgumentSInt8(ArgumentNumBase):
    @property
    def _type_info(self) -> int:
        return TypeInfo.TYPE_SIGNED | TypeInfo.TYPE_LENGTH_8BIT

    @staticmethod
    def _struct_format() -> str:
        return "b"

    @staticmethod
    def _data_payload_length() -> int:
        return 1

    @property
    def data_payload_length(self) -> int:
        return self._data_payload_length()


class ArgumentSInt16(ArgumentNumBase):
    @property
    def _type_info(self) -> int:
        return TypeInfo.TYPE_SIGNED | TypeInfo.TYPE_LENGTH_16BIT

    @staticmethod
    def _struct_format() -> str:
        return "h"

    @staticmethod
    def _data_payload_length() -> int:
        return 2

    @property
    def data_payload_length(self) -> int:
        return self._data_payload_length()


class ArgumentSInt32(ArgumentNumBase):
    @property
    def _type_info(self) -> int:
        return TypeInfo.TYPE_SIGNED | TypeInfo.TYPE_LENGTH_32BIT

    @staticmethod
    def _struct_format() -> str:
        return "i"

    @staticmethod
    def _data_payload_length() -> int:
        return 4

    @property
    def data_payload_length(self) -> int:
        return self._data_payload_length()


class ArgumentSInt64(ArgumentNumBase):
    @property
    def _type_info(self) -> int:
        return TypeInfo.TYPE_SIGNED | TypeInfo.TYPE_LENGTH_64BIT

    @staticmethod
    def _struct_format() -> str:
        return "q"

    @staticmethod
    def _data_payload_length() -> int:
        return 8

    @property
    def data_payload_length(self) -> int:
        return self._data_payload_length()


class ArgumentFloat32(ArgumentNumBase):
    @property
    def _type_info(self) -> int:
        return TypeInfo.TYPE_FLOAT | TypeInfo.TYPE_LENGTH_32BIT

    @staticmethod
    def _struct_format() -> str:
        return "f"

    @staticmethod
    def _data_payload_length() -> int:
        return 4

    @property
    def data_payload_length(self) -> int:
        return self._data_payload_length()


class ArgumentFloat64(ArgumentNumBase):
    @property
    def _type_info(self) -> int:
        return TypeInfo.TYPE_FLOAT | TypeInfo.TYPE_LENGTH_64BIT

    @staticmethod
    def _struct_format() -> str:
        return "d"

    @staticmethod
    def _data_payload_length() -> int:
        return 8

    @property
    def data_payload_length(self) -> int:
        return self._data_payload_length()


class ArgumentByteBase(Argument):
    """It is a class for byte-base argument.

    It has a length field as 4 byte and data field for value.
    """

    LENGTH_SIZE = 2

    @property
    def data_payload_length(self) -> int:
        return self.LENGTH_SIZE + self.data_length

    @property
    @abstractmethod
    def data_length(self) -> int:
        raise NotImplementedError

    def data_payload_to_bytes(self, msb_first: Optional[bool] = None) -> bytes:
        if msb_first is not None:
            endian = ">" if msb_first else "<"
        elif self.msb_first is not None:
            endian = ">" if self.msb_first else "<"
        elif msb_first_default is not None:
            endian = ">" if msb_first_default else "<"
        else:
            raise ValueError("Endian is not known")
        return struct.pack(f"{endian}H", self.data_length) + self.data_to_bytes()

    @abstractmethod
    def data_to_bytes(self) -> bytes:
        raise NotImplementedError


class ArgumentString(ArgumentByteBase):
    def __init__(
        self,
        data: str,
        is_utf8: bool = False,
        msb_first: Optional[bool] = None,
        encoding: Optional[str] = None,
    ):
        """Create argument of Type String.

        Args:
            data (str): A data payload of string.
            is_utf8 (bool, optional): Encoding of the string is UTF-8 if True, or ASCII.
                                      Defaults to True (UTF-8).
            msb_first (Optional[bool], optional): [description]. Defaults to None.
            encoding: custom 8-bit encoding that will be used for serialization
                      Has no effect if is_utf8 is set to True
        """
        super().__init__(msb_first)
        self.data = data
        self.is_utf8 = is_utf8
        self._encoding = self._encoding_format(is_utf8, encoding)

    def _to_str(self) -> str:
        return self.data

    @property
    def _type_info(self) -> int:
        type_info = TypeInfo.TYPE_STRING
        if self.is_utf8:
            type_info |= TypeInfo.STRING_CODING_UTF8
        return type_info

    @classmethod
    def from_data_payload(
        cls,
        data_payload: bytes,
        is_utf8: bool,
        msb_first: bool,
        encoding: Optional[str] = None,
    ) -> "Argument":
        endian = ">" if msb_first else "<"
        length = struct.unpack(f"{endian}H", data_payload[: cls.LENGTH_SIZE])[0]
        return cls(
            data_payload[cls.LENGTH_SIZE : cls.LENGTH_SIZE + length - 1].decode(
                cls._encoding_format(is_utf8, encoding), "replace"
            ),
            is_utf8,
            msb_first,
        )

    @property
    def data_length(self) -> int:
        return len(self.data.encode(self._encoding, "replace")) + 1

    def data_to_bytes(self) -> bytes:
        return self.data.encode(self._encoding, "replace") + b"\x00"

    @staticmethod
    def _encoding_format(is_utf8: bool, encoding: str) -> str:
        if encoding is None:
            encoding = "ascii"
        return "utf-8" if is_utf8 else encoding


class ArgumentRaw(ArgumentByteBase):
    def __init__(
        self,
        data: bytes,
        msb_first: Optional[bool] = None,
    ):
        super().__init__(msb_first)
        self.data = data

    def _to_str(self) -> str:
        return self.data.hex()

    @property
    def _type_info(self) -> int:
        return TypeInfo.TYPE_RAW

    @classmethod
    def from_data_payload(cls, data_payload: bytes, msb_first: bool) -> "Argument":
        endian = ">" if msb_first else "<"
        length = struct.unpack(f"{endian}H", data_payload[: cls.LENGTH_SIZE])[0]
        return cls(data_payload[cls.LENGTH_SIZE : cls.LENGTH_SIZE + length], msb_first)

    @property
    def data_length(self) -> int:
        return len(self.data)

    def data_to_bytes(self) -> bytes:
        return self.data


class VerbosePayload(Payload):
    """The Payload of Verbose Mode of a DLT Message."""

    def __init__(self, arguments: List[Argument]) -> None:
        """Create VerbosePayload object.

        Args:
            arguments (List[Argument]): A list of argument
        """
        self.arguments = arguments

    @classmethod
    def create_from_bytes(
        cls,
        data: bytes,
        msb_first: bool,
        number_of_arguments: int,
        encoding: Optional[str] = None,
    ) -> "VerbosePayload":
        """Create VerbosePayload object from data bytes.

        Args:
            data (bytes): Data bytes
            msb_first (bool): If set, the payload data is in big endian format,
                              else in little endian format.
            number_of_arguments: number of arguments within this payload data
            encoding: optional non-standard 8-bit string encoding

        Raises:
            ValueError: It can be caused by invalid data format.

        Returns:
            VerbosePayload: New VerbosePayload object
        """
        arguments = []
        offset = 0
        encodings = [
            'utf-8',
            'latin-1',
            'cp1252',
            'ascii',
            'utf-16',
            'utf-32',
            'iso-8859-1',
            'gb2312',
            'gbk',
            'shift_jis',
            'euc-kr',
        ]
        encodings = [
            'utf-8',
            'latin-1',
            'cp1252',
            'ascii',
            'utf-16',
            'utf-32',
            'iso-8859-1',
            'iso-8859-2',
            'iso-8859-3',
            'iso-8859-4',
            'iso-8859-5',
            'iso-8859-6',
            'iso-8859-7',
            'iso-8859-8',
            'iso-8859-9',
            'iso-8859-10',
            'iso-8859-11',
            'iso-8859-13',
            'iso-8859-14',
            'iso-8859-15',
            'iso-8859-16',
            'windows-1250',
            'windows-1251',
            'windows-1252',
            'windows-1253',
            'windows-1254',
            'windows-1255',
            'windows-1256',
            'windows-1257',
            'windows-1258',
            'us-ascii',
            'gb2312',
            'gbk',
            'gb18030',
            'big5',
            'shift_jis',
            'euc-jp',
            'euc-kr',
            'koi8-r',
            'koi8-u',
            'latin-greek',
            'latin5',
            'iso-8859-11',
            'macintosh',
            'mac_cyrillic',
            'mac_greek',
            'mac_iceland',
            'mac_latin2',
            'mac_roman',
            'mac_turkish',
            'mac_ukraine',
            'ptcp154',
            'cp437',
            'cp737',
            'cp775',
            'cp850',
            'cp852',
            'cp855',
            'cp857',
            'cp858',
            'cp860',
            'cp861',
            'cp862',
            'cp863',
            'cp864',
            'cp865',
            'cp866',
            'cp869',
            'hp_roman8',
            'tis-620',
            'iso-2022-jp',
            'iso-2022-kr',
            'iso-2022-cn',
            'iso-8859-6-e',
            'iso-8859-6-i',
            'iso-8859-8-e',
            'iso-8859-8-i',
            'iso-2022-jp-1',
            'iso-2022-jp-2',
            'iso-2022-jp-3',
            'iso-2022-jp-ext',
            'iso-2022-cn-cns',
            'iso-2022-cn-gb',
            'iso-2022-kr-7',
            'iso-2022-kr-7bit',
            'hz-gb-2312',
            'euc-jis-2004',
            'iso-ir-165',
            'iso-646-irv',
            'iso-10646-ucs-2',
            'iso-10646-ucs-4',
            'unicode-internal',
            'x-user-defined',
            'replacement',
        ]
        for _ in range(number_of_arguments):
            arg = None
            for enc in encodings:
                try:
                    arg = Argument.create_from_bytes(data[offset:], msb_first, encoding)
                except:
                    pass
                if arg:
                    break
            if arg:
                arguments.append(arg)
                offset += arg.bytes_length
        return cls(arguments)

    def to_bytes(self, msb_first: Optional[bool] = None) -> bytes:
        """Convert to data bytes.

        Args:
            msb_first:  True - Handle payload data as big endian
                       False - Handle payload data as little endian
                        None - Handle payload data according to field value of endian
                               type. The value of msb_first_default is used
                               if field value of endian type is also None.
                               Raise ValueError if both are None.

        Raises:
            ValueError: All endian type of the argument, field value,
                        and msb_first_default are None

        Returns:
            bytes: Converted data bytes
        """
        data = b""
        for arg in self.arguments:
            data += arg.to_bytes(msb_first)
        return data

    @property
    def bytes_length(self) -> int:
        """Get length of the data bytes.

        Returns:
            int: Length of the data bytes
        """
        length = 0
        for arg in self.arguments:
            length += arg.bytes_length
        return length

    def _to_str(self) -> str:
        """Convert payload to human readable string.

        Returns:
            str: Human readable string.
        """
        return " ".join([str(arg) for arg in self.arguments])
